@@ -1,15 +1,15 @@
 
 /* Support for dynamic loading of extension modules */
 
+#include "Python.h"
+#include "importdl.h"
+
 #include <ctype.h>	/*  for isdigit()	  */
 #include <errno.h>	/*  for global errno      */
 #include <string.h>	/*  for strerror()        */
 #include <stdlib.h>	/*  for malloc(), free()  */
 #include <sys/ldr.h>
 
-#include "Python.h"
-#include "importdl.h"
-
 
 #ifdef AIX_GENUINE_CPLUSPLUS
 #include "/usr/lpp/xlC/include/load.h"
@@ -103,19 +103,6 @@ aix_getoldmodules(void **modlistptr)
 	return 0;
 }
 
-static int
-aix_bindnewmodule(void *newmoduleptr, void *modlistptr)
-{
-	register ModulePtr modptr;
-
-	/*
-	-- Bind the new module with the list of loaded modules.
-	*/
-	for (modptr = (ModulePtr)modlistptr; modptr; modptr = modptr->next)
-		if (loadbind(0, modptr->entry, newmoduleptr) != 0)
-			return -1;
-	return 0;
-}
 
 static void
 aix_loaderror(const char *pathname)
@@ -157,7 +144,7 @@ aix_loaderror(const char *pathname)
 		    if (nerr == load_errtab[j].errNo && load_errtab[j].errstr)
 			ERRBUF_APPEND(load_errtab[j].errstr);
 		}
-		while (isdigit(*message[i])) message[i]++ ; 
+		while (isdigit(Py_CHARMASK(*message[i]))) message[i]++ ; 
 		ERRBUF_APPEND(message[i]);
 		ERRBUF_APPEND("\n");
 	}
@@ -191,10 +178,6 @@ dl_funcptr _PyImport_GetDynLoadFunc(const char *fqname, const char *shortname,
 		aix_loaderror(pathname);
 		return NULL;
 	}
-	if (aix_bindnewmodule((void *)p, staticmodlistptr) == -1) {
-		aix_loaderror(pathname);
-		return NULL;
-	}
 
 	return p;
 }
